from functools import cached_property

from flask_wtf import FlaskForm
from sqlalchemy import select
from wtforms.fields import BooleanField, StringField, TextAreaField
from wtforms.fields.simple import SubmitField
from wtforms.validators import DataRequired, Length, Optional

from labbase2.database import db
from labbase2.forms import filters, rendering
from labbase2.models import Group, Permission, User

__all__ = ["AddGroupForm", "EditGroupForm", "ChangeGroupsForm"]


class ChangeGroupsForm(FlaskForm):

    submit = SubmitField(label="Submit", render_kw=rendering.submit_field)

    def __init__(self, user_id: int, *args, **kwargs):
        super().__init__(*args, **kwargs)

        user = db.session.get(User, user_id)

        # Add a BooleanField for each permission in the db.
        for group in db.session.scalars(
            select(Group).where(Group.name.notin_(["admin", "user"]))
        ).all():
            field_name = f"g_{group.name}"
            field = BooleanField(
                group.name,
                description=group.description,
                default=group in user.groups,
            )
            bound_field = self.meta.bind_field(
                self, field, {"name": field_name, "prefix": self._prefix}
            )
            bound_field.process(None)
            setattr(self, field_name, bound_field)
            self._fields[field_name] = bound_field

    @cached_property
    def selected_groups(self) -> list[Group]:
        groups = [key[2:] for key, value in self.data.items() if key.startswith("g_") and value]
        return db.session.scalars(select(Group).where(Group.name.in_(groups))).all()


class EditGroupForm(FlaskForm):

    description = TextAreaField(
        label="Description",
        validators=[Optional()],
        filters=[filters.strip_input],
        render_kw=rendering.custom_field | {"placeholder": "Some useful description...", "size": 5},
    )
    submit = SubmitField(label="Submit", render_kw=rendering.submit_field)

    def __init__(self, group_name: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if group_name:
            group = db.session.get(Group, group_name)

        # Add a BooleanField for each permission in the db.
        for permission in db.session.scalars(select(Permission)).all():
            field_name = f"p_{permission.name}".replace("-", "_")
            field = BooleanField(
                permission.name,
                description=permission.description,
                default=False if not group_name else permission in group.permissions,
            )
            bound_field = self.meta.bind_field(
                self, field, {"name": field_name, "prefix": self._prefix}
            )
            bound_field.process(None)
            setattr(self, field_name, bound_field)
            self._fields[field_name] = bound_field

    @cached_property
    def selected_permissions(self) -> list[Permission]:
        permissions = [
            key[2:].replace("_", "-")
            for key, value in self.data.items()
            if key.startswith("p_") and value
        ]
        return db.session.scalars(select(Permission).where(Permission.name.in_(permissions))).all()


class AddGroupForm(EditGroupForm):

    name = StringField(
        label="Name",
        validators=[DataRequired(), Length(max=32)],
        filters=[filters.strip_input, filters.make_lower],
        render_kw=rendering.custom_field | {"placeholder": "Group name"},
        description="A unique name for the group. Choose wisely as you cannot change this name later on.",
    )



__all__ = ["DefaultConfig"]


class DefaultConfig:
    """A default config for logging"""

    SECRET_KEY: str = "verysecretkey"

    # Database.
    SQLALCHEMY_DATABASE_URI: str = "sqlite:///labbase2.db"
    SQLALCHEMY_TRACK_MODIFICATIONS: bool = False

    # Display.
    DELETABLE_HOURS: int = 72
    PER_PAGE: int = 100

    # User.
    USER: tuple = "Admin", "Admin", "admin@admin.de"
    DEFAULT_TIMEZONE: str = "Europe/Berlin"
    UPLOAD_FOLDER: str = "upload"
    RESET_EXPIRES: int = 10
    PERMISSIONS: list[tuple[str, str]] = [
        ("add-comment", "Allows a user to author own comments."),
        ("upload-file", "Allows a user to upload files."),
        ("add-dilution", "Allows a user to add dilutions to antibodies."),
        ("add-preparation", "Allows a user to add plasmid preparations."),
        ("add-glycerol-stock", "Allows a user to create glycerol stocks."),
        ("add-consumable-batch", "Allows a user to add batches of consumables."),
        ("add-antibody", "Allows a user to add antibodies."),
        ("delete-antibody", "Allow a user to delete antibodies."),
        ("add-chemical", "Allows a user to add chemicals."),
        ("add-stock-solution", "Allows a user to add stock solutions of chemicals."),
        ("add-fly-stock", "Allows a user to create fly stocks."),
        ("add-plasmid", "Allows a user to add cloned plasmids."),
        ("add-oligonucleotide", "Allows a user to add primer/oligonucleotides."),
        ("view-user", "Allows a user to see all other users."),
        ("inactivate-user", "Allow a user to inactivate a user account."),
        ("reset-password", "Allow a user to reset passwords for everyone."),
        ("register-user", "Allows a user to register another user."),
        ("change-group", "Allow a user to assign groups to users"),
        ("edit-groups", "Allow a user to create new groups and assign permissions to groups."),
        ("delete-user", "Allows a user to delete another user."),
        ("export-content", "Allows a user to export content. Suggested level: PhD Student."),
        ("add-request", "Allows a user to add requests for any ressource. Suggested level: PI."),
    ]

    # Data.
    RESISTANCES: list[str] = [
        "Ampicillin",
        "Ampicillin and Kanamycin",
        "Apramycin",
        "Chloramphenicol",
        "Chloramphenicol and Ampicillin",
        "Gentamicin",
        "Kanamycin",
        "Spectinomycin",
        "Streptomycin",
        "Tetracyclin",
    ]
    STRAINS: list[str] = [
        "ccdB Survival 2 T1",
        "DB3.1",
        "DH10B",
        "DH5alpha",
        "HB101",
        "JM109",
        "JM110",
        "MC1061",
        "MG1655",
        "NEB Stable",
        "Pir1",
        "Stbl2",
        "Stbl3",
        "Stbl4",
        "TOP10",
        "XL1 Blue",
        "XL10 Gold",
    ]



from sqlalchemy import event, func

from labbase2.database import db
from labbase2.models import ColumnMapping, Group, Oligonucleotide, User, file


@event.listens_for(db.session, "deleted_to_detached")
def intercept_deleted_to_detached(session, obj) -> None:
    """Removes the physical file when a File row marked as deleted is
    eventually removed from the database.

    Parameters
    ----------
    session
        The current session.
    obj
        The object that is detached. This can be any Database related object
        but so far this function has only implications for 'File'.


    Returns
    -------
    None
    """

    print("Deleting file.")

    if isinstance(obj, file.BaseFile):
        obj.path.unlink(missing_ok=True)


# TODO: There must be a better option than writing an event for every single child
#  table of BaseEntity.
@event.listens_for(Oligonucleotide, "before_update")
def update_parent(mapper, connection, target) -> None:
    target.timestamp_edited = func.now()


@event.listens_for(ColumnMapping, "before_update")
def update_import_job(mapper, connection, target) -> None:
    target.job.timestamp_edited = func.now()


@event.listens_for(db.session, "before_flush")
def add_default_user_group(session, flush_context, instances):
    """Add the group `user` to every new user automatically

    Every new user is automatically member of the `user` group. This can not be changed.
    In order to avoid errors, the `user` group should not be added to a user
    within views or any other place.

    Parameters
    ----------
    session
    flush_context
    instances

    Returns
    -------
    None
    """

    user_group = db.session.get(Group, "user")

    for obj in session.new:
        if isinstance(obj, User) and user_group not in obj.groups:
            obj.groups.append(user_group)


@event.listens_for(Group, "before_delete")
def protect_standard_groups(mapper, connection, target):
    if target.name in ["admin", "user"]:
        raise ValueError("Groups 'admin' and 'user' cannot be deleted!")


@event.listens_for(User.groups, "remove")
def protect_user_membership(user, group, initiator):
    if group.name == "user":
        raise ValueError("Group 'user' cannot be removed from a user!")



import json
import secrets
from pathlib import Path
from typing import Optional, Union

from flask import Flask
from sqlalchemy import func, select

from labbase2 import logging, views
from labbase2.database import db
from labbase2.models import Group, Permission, User, events
from labbase2.models.user import login_manager
from labbase2.utils import template_filters

__all__ = ["create_app"]


def create_app(
    config: Optional[Union[str, Path]] = None,
    config_dict: Optional[dict] = None,
    **kwargs,
) -> Flask:
    """Create an app instance of the labbase2 application.

    Parameters
    ----------
    config : Optional[Union[str, Path]]
        A filename pointing to the configuration file. File has to be in JSON format.
        Filename is supposed to be relative to the instance path.
    config_dict : Optional[dict]
        Additional config parameters for the app. If `config` und `config_dict` contain
        the same keys, settings from `config_dict` will be applied.
    kwargs
        Additional parameters passed to the Flask class during instantiation. Supports
        all parameters of the Flask class except `import_name` and
        `instance_relative_config`, which are hardcoded to `labbase2` and `True`
        respectively.

    Returns
    -------
    Flask
        A configured Flask application instance. If run for the first time, an instance
        folder as well as a sub-folder for uploading files and a SQLite database will be
        created.
    """

    app: Flask = Flask("labbase2", instance_relative_config=True, **kwargs)
    app.config.from_object("labbase2.config.DefaultConfig")

    if config is not None:
        app.config.from_file(config, load=json.load, text=False)
    if config_dict is not None:
        app.config |= config_dict

    app.logger.debug("Config dict: %s", config_dict)
    app.logger.debug("Instance path: %s", app.instance_path)

    # Initialize logging.
    logging.init_app(app)

    # Create upload folder if necessary.
    try:
        Path(app.instance_path, app.config["UPLOAD_FOLDER"]).mkdir(parents=True, exist_ok=True)
    except PermissionError as error:
        app.logger.error("Could not create upload folder due to insufficient permissions!")
        raise error

    # Initiate the database.
    db.init_app(app)

    with app.app_context():
        # Create database and add tables (if not yet present).
        db.create_all()

    # Create/update permissions from the config.
    _set_up_permissions(app=app)
    _set_up_groups(app=app)

    # If no user with admin rights is in the database, create one.
    _set_up_admin(app=app)

    # Register login_manager with application.
    login_manager.init_app(app)

    # Register blueprints with application.
    app.register_blueprint(views.base.bp)
    app.register_blueprint(views.auth.bp)
    app.register_blueprint(views.imports.bp)
    app.register_blueprint(views.chemicals.bp)
    app.register_blueprint(views.comments.bp)
    app.register_blueprint(views.files.bp)
    app.register_blueprint(views.fly_stocks.bp)
    app.register_blueprint(views.requests.bp)
    app.register_blueprint(views.batches.bp)
    app.register_blueprint(views.antibodies.bp)
    app.register_blueprint(views.plasmids.bp)
    app.register_blueprint(views.oligonucleotides.bp)

    # Add custom template filters to Jinja2.
    app.jinja_env.filters["format_date"] = template_filters.format_date
    app.jinja_env.filters["format_datetime"] = template_filters.format_datetime
    app.jinja_env.globals["random_string"] = secrets.token_hex

    return app


def _set_up_permissions(app: Flask):
    with app.app_context():
        # Add permissions to database.
        for name, description in app.config.get("PERMISSIONS"):
            if (permission := db.session.get(Permission, name)) is None:
                db.session.add(Permission(name=name, description=description))
            else:
                permission.description = description

        db.session.commit()


def _set_up_groups(app: Flask):
    with app.app_context():
        if not db.session.get(Group, "admin"):
            group_admin = Group(name="admin")
            db.session.add(group_admin)
        else:
            group_admin = db.session.get(Group, "admin")

        group_admin.permissions = db.session.scalars(select(Permission)).all()
        db.session.commit()

        if not db.session.get(Group, "user"):
            group_user = Group(name="user")
            db.session.add(group_user)
        else:
            group_user = db.session.get(Group, "user")

        group_user.permissions = [db.session.get(Permission, "add-comment")]
        db.session.commit()

        # Make sure, every user is in group "user".
        for user in db.session.scalars(select(User)).all():
            if not group_user in user.groups:
                user.groups.append(group_user)

        db.session.commit()


def _set_up_admin(app: Flask):
    with app.app_context():
        first, last, email = app.config.get("USER")

        user_count = select(func.count()).select_from(User)  # pylint: disable=not-callable
        admin_count = (
            select(func.count())  # pylint: disable=not-callable
            .select_from(User)
            .join(User.groups)
            .where(User.is_active, Group.name == "admin")
        )

        if db.session.scalar(user_count) == 0:
            app.logger.info("No user in database; create admin as specified in config.")
            admin = User(first_name=first, last_name=last, email=email)
            admin.set_password("admin")
            db.session.add(admin)
            db.session.commit()
            admin.groups.append(db.session.get(Group, "admin"))
            # admin.groups.append(db.session.get(Group, "user"))
            db.session.commit()
        elif db.session.scalar(admin_count) == 0:
            app.logger.info("No active user with admin rights; Re-activate admin.")
            admin = db.session.scalars(select(User).where(User.email == email)).first()
            if admin is not None:
                app.logger.info("Re-activated admin: %s", admin.username)
                admin.groups.append(db.session.get(Group, "admin"))
            else:
                app.logger.info("Did not find admin, create an admin from config.")
                admin = User(first_name=first, last_name=last, email=email, is_admin=True)
                admin.set_password("admin")
                admin.groups.append(db.session.get(Group, "admin"))
                # admin.groups.append(db.session.get(Group, "user"))
                db.session.add(admin)

        try:
            db.session.commit()
        except Exception as error:
            app.logger.error("Could not add initial user/admin to database: %s", error)
            raise error



import logging

from flask import Flask, has_request_context, request
from flask_login import AnonymousUserMixin
from flask_login import current_user as user

__all__ = ["init_app", "RequestFormatter"]


def init_app(app: Flask):
    """Initialize logging with the current app

    Parameters
    ----------
    app: Flask
        A flask app to initialize.

    Returns
    -------
    None
    """

    @app.before_request
    def auto_log_request():
        if not request.path.startswith("/static") and not request.path.startswith("/favicon"):
            app.logger.info(
                "%(method)-6s %(path)s", {"method": request.method, "path": request.url}
            )


class RequestFormatter(logging.Formatter):
    """Customized logging formatter to inject flask request information"""

    def format(self, record: logging.LogRecord):
        """Inject request information to logging messages and format that message

        Parameters
        ----------
        record: logging.LogRecord
            A LogRecord instance.

        Returns
        -------
        str
            A formatted log message.
        """

        if has_request_context():
            record.url = request.url
            record.remote_addr = request.remote_addr
            record.user = "Anonymous" if isinstance(user, AnonymousUserMixin) else user.username
        else:
            record.url = None
            record.remote_addr = None
            record.user = "Anonymous"

        return super().format(record)



from flask import url_for
from sqlalchemy import func, select

from labbase2 import models
from labbase2.database import db


def test_first_user_is_admin_and_active(app):
    with app.app_context():
        user = db.session.get(models.User, 1)
        group = db.session.get(models.Group, "admin")

        assert user is not None
        assert group in user.groups
        assert user.is_active
        assert user.first_name == "Max"
        assert user.last_name == "Mustermann"
        assert user.email == "test@test.de"


def test_all_permissions_in_db(app):
    with app.app_context():
        permissions_count = db.session.scalar(select(func.count()).select_from(models.Permission))

    assert permissions_count == len(app.config["PERMISSIONS"])


def test_first_user_has_all_permissions(app):
    with app.app_context():
        permissions_count = db.session.scalar(select(func.count()).select_from(models.Permission))
        user = db.session.get(models.User, 1)

        admin_group = db.session.get(models.Group, "admin")
        user_group = db.session.get(models.Group, "user")

    assert permissions_count == len(admin_group.permissions)
    assert admin_group in user.groups
    assert user_group in user.groups



import pandas as pd
from sqlalchemy import inspect, select
from sqlalchemy.orm import Mapped

from labbase2.database import db

__all__ = ["Importer"]


class Importer:

    import_attr: tuple = (("id", "ID"),)
    not_updatable: tuple = ("id",)

    def update(self, **kwargs) -> None:
        """Update attributes of an entity.

        Parameters
        ----------
        kwargs
            Attributes to be updated and the new value.

        Returns
        -------
        None

        Notes
        -----
        The changes are not committed automatically to the database, i.e.,
        any changes done via update will not be saved if the calling code
        does not commit these changes on its own.
        """

        for attr, value in kwargs.items():
            if hasattr(self, attr):
                setattr(self, attr, value)

    @classmethod
    def importable_fields(cls) -> list[Mapped]:
        """Get a list of all importable columns

        Returns
        -------
        list[Mapped]
            A list of mapped columns that have been tagged as importable.
        """

        fields = []
        for column in inspect(cls).columns:
            if column.info.get("importable", False):
                fields.append(column.name)
        return fields

    @classmethod
    def from_record(cls, rec: dict, update: bool = False):
        """Create a class instance from a dict

        Parameters
        ----------
        rec: dict
        update: bool

        Returns
        -------
        cls
            An instance of this class
        """

        rec = cls.process_record(rec=rec)

        if update:
            if id_ := rec.pop("id", None):
                entity = db.session.get(cls, id_)
            elif label := rec.pop("label", None):
                entity = db.session.scalar(select(cls).where(cls.label == label))
            else:
                entity = None

            if not entity:
                return None

            for key in rec:
                if key in cls.not_updatable:
                    del rec[key]

            entity.update(**rec)

        else:
            if "id" in rec:
                del rec["id"]

            entity = cls(**rec)

        return entity

    @classmethod
    def process_record(cls, rec: dict) -> dict:
        """Process a row of pd.DataFrame represented as a dict

        Parameters
        ----------
        rec: dict

        Returns
        -------
        dict
            A dict with all entries remove where the value returns `True` for
            `pd.isnull`.
        """

        return {k: v for k, v in rec.items() if not pd.isnull(v)}

    # @classmethod
    # def import_form(cls, columns: list, *args, **kwargs) -> ImportEntity:
    #     data = {'mappings': len(columns) * [[]]}
    #
    #     form = ImportEntity(clss=cls.__name__, data=data, *args, **kwargs)
    #
    #     for column, field in zip(columns, form.mappings):
    #         field.label = column
    #         field.choices += cls.import_attr
    #
    #     return form

    @staticmethod
    def process_formdata(data: dict) -> dict:
        return {k: v for k, v in data.items() if v}



from datetime import datetime

from sqlalchemy import DateTime, ForeignKey, func
from sqlalchemy.orm import Mapped, mapped_column, relationship

from labbase2.database import db

__all__ = ["ImportJob", "ColumnMapping"]


class ImportJob(db.Model):
    """A class to hold information for importing a file

    Attributes
    ----------
    id: int
    user_id: int
    timestamp: datetime
    timestamp_edited: datetime
    file_id: int
    is_finished: bool
    entity_type: str
    mappings: list[ColumnMapping]
    file: BaseFile
    """

    __tablename__ = "import_job"

    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("user.id"), nullable=False)
    timestamp: Mapped[datetime] = mapped_column(
        DateTime,
        nullable=False,
        default=func.now(),  # pylint: disable=not-callable
    )
    timestamp_edited: Mapped[datetime] = mapped_column(
        DateTime,
        nullable=True,
        default=func.now(),  # pylint: disable=not-callable
    )
    file_id: Mapped[int] = mapped_column(ForeignKey("base_file.id"), nullable=False)
    is_finished: Mapped[bool] = mapped_column(default=False, nullable=False)
    entity_type: Mapped[str] = mapped_column(nullable=False)

    # One-to-many relationships.
    mappings: Mapped[list["ColumnMapping"]] = relationship(
        backref="job", cascade="all, delete-orphan", lazy=True
    )
    file: Mapped["BaseFile"] = relationship(
        backref="import_job", lazy=True, cascade="all, delete", single_parent=True
    )

    def get_file(self):
        pass


class ColumnMapping(db.Model):
    """A class to represent the mapping between a file and database columns

    Attributes
    job_id: int
    mapped_field: str
        The name of the column in the database.
    input_column: str
        The name of the column in the import file.
    """

    __tablename__ = "column_mapping"

    job_id: Mapped[int] = mapped_column(ForeignKey("import_job.id"), primary_key=True)
    mapped_field: Mapped[str] = mapped_column(primary_key=True)
    input_column: Mapped[str] = mapped_column(nullable=True)



import io
from datetime import date

import pandas as pd
from flask import send_file
from sqlalchemy import inspect

from labbase2.database import db

__all__ = ["Export"]


class Export:
    """A mixin for exporting instances to CSV, JSON, ...."""

    def to_dict(self) -> dict:
        """Create a dictionary from the object.

        Returns
        -------
        dict
            A dictionary. The keys are exactly labeled like the attributes of the
            instance.
        """

        inst = inspect(self).mapper.column_attrs
        return {c.key: getattr(self, c.key) for c in inst}

    @classmethod
    def to_df(cls, instances: list[db.Model]) -> pd.DataFrame:
        """Create a pandas dataframe from a list of instances

        Parameters
        ----------
        instances: list[db.Model]

        Returns
        -------
        pd.DataFrame
            A pandas dataframe. Each row reprsents one row in the database, which
            usually consists of joined tables to fully represent the respective
            instance.
        """

        instances = db.session.scalars(instances)
        return pd.DataFrame(i.to_dict() for i in instances)

    @classmethod
    def _filename(cls) -> str:
        return cls.__name__ + "_" + date.today().isoformat()

    @classmethod
    def export_to_csv(cls, instances: list[db.Model]):
        """Export a list of database model instances to CSV

        Parameters
        ----------
        instances: list[db.Model]


        Returns
        -------
        None
        """

        with io.StringIO() as proxy:
            cls.to_df(instances).to_csv(proxy)
            mem = io.BytesIO(proxy.getvalue().encode("utf-8"))

        return send_file(
            mem,
            as_attachment=True,
            download_name=cls._filename() + ".csv",
            mimetype="text/csv",
        )

    @classmethod
    def export_to_json(cls, instances):
        """Export a list of database model instances to JSON

        Parameters
        ----------
        instances: list[db.Model]


        Returns
        -------
        None
        """

        with io.StringIO() as proxy:
            cls.to_df(instances).to_json(proxy, orient="records", date_format="iso", indent=2)
            mem = io.BytesIO(proxy.getvalue().encode("utf-8"))

        return send_file(
            mem,
            as_attachment=True,
            download_name=cls._filename() + ".json",
            mimetype="text/json",
        )

    @classmethod
    def to_pdf(cls, instances):
        """Export a list of database model instances to PDF

        Parameters
        ----------
        instances: list[db.Model]


        Returns
        -------
        None

        Notes
        -----
        This method is not implemented yet. Nevertheless, it returns an empty file.
        """

        mem = io.BytesIO()
        mem.seek(0)

        for instance in instances:
            pass

        return send_file(
            mem,
            as_attachment=True,
            download_name=cls._filename() + ".pdf",
            mimetype="application/pdf",
        )



from flask import render_template
from flask_wtf import FlaskForm
from wtforms.fields import BooleanField, Field, IntegerField, SelectField, StringField, SubmitField
from wtforms.validators import NumberRange, Optional

from . import filters, rendering

__all__ = ["FilterForm", "EditEntityForm"]


class FilterForm(FlaskForm):
    """Base class for filter forms

    Attributes
    ----------
    id: IntegerField
    ascending: BooleanField
    order_by: SelectField
    submit: SelectField
    download_csv: SelectField
    download_pdf: SelectField
    download_excel: SelectField
    """

    id = IntegerField(
        label="ID",
        validators=[Optional(), NumberRange(min=1)],
        render_kw=rendering.custom_field | {"placeholder": "ID"},
        description="Internal database ID.",
    )
    ascending = BooleanField(
        label="Sort ascending",
        render_kw=rendering.boolean_field,
        default=True,
        description="Uncheck to sort results in descending order.",
    )
    order_by = SelectField(
        label="Order by",
        choices=[("id", "ID")],
        default="id",
        render_kw=rendering.select_field,
        description="The column by which the results shall be ordered.",
    )
    submit = SubmitField(label="Submit", render_kw=rendering.submit_field)
    download_csv = SubmitField(label="Export to CSV", render_kw=rendering.submit_field)
    download_pdf = SubmitField(label="Export to PDF", render_kw=rendering.submit_field)
    download_excel = SubmitField(label="Export to Excel", render_kw=rendering.submit_field)

    def fields(self) -> list[Field]:
        """Returns all fields for rendering"""

        raise NotImplementedError

    def render(self, action: str = "", method: str = "GET") -> str:
        """Render the form to HTML

        Parameters
        ----------
        action: str, optional
            The target URL for the form.
        method: str, optional
            The request method, e.g. 'GET' or 'POST'.

        Returns
        -------
        str
            The rendered HTML.
        """

        return render_template("forms/filter.html", form=self, action=action, method=method)


class EditEntityForm(FlaskForm):
    """The base form to edit entries in the database.

    Attributes
    ----------
    label: StringField
    submit: SubmitField
    """

    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[filters.strip_input],
        render_kw=rendering.custom_field | {"placeholder": "Name"},
        description="Must be unique among ALL database entries.",
    )
    submit = SubmitField(label="Submit", render_kw=rendering.submit_field)



from datetime import date

from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from labbase2.database import db
from labbase2.models.base_entity import BaseEntity
from labbase2.models.fields import CustomDate
from labbase2.models.mixins.importer import Importer

__all__ = ["Modification", "FlyStock"]


class Modification(db.Model, Importer):
    """
    Information about modification of fly stocks. This should be things
    like re-establishing a dead fly stock. If the genotype is altered it
    might be rather better to create a new fly stock entry.

    Attributes
    ----------
    id : int
        An internal identifier for this modifcation.
    fly_id : int
        The identifier of the fly stock to which this modifcation belongs.
    user_id : int
        ID of the person that did the modification.
    date : date
        Date of the modification. Defaults to the current date.
    description : str
        A short and precise explanation of what was done.

    """

    id: Mapped[int] = mapped_column(primary_key=True)
    fly_id: Mapped[int] = mapped_column(ForeignKey("fly_stock.id"), nullable=False)
    user_id: Mapped[int] = mapped_column(ForeignKey("user.id"), nullable=False)
    date: Mapped[date] = mapped_column(db.Date, nullable=True)
    description: Mapped[str] = mapped_column(String(1024), nullable=True)


class FlyStock(BaseEntity):
    """A fly stock.

    Attributes
    ----------
    id : int
        The internal ID of this fly stock. This ID is unique among ALL entities
        in the database.
    short_genotype : str
        The complete genotype of this stock. Heterologeous chromosomes are
        separated by ';' and homologeous chromosomes are separated by '/'.
        The order of the chromosomes is 'x ; y ; 2 ; 3 ; 4'.
    chromosome_xa : str
    chromosome_xb : str
    chromosome_y : str
    chromosome_2a : str
    chromosome_2b : str
    chromosome_3a : str
    chromosome_3b : str
    chromosome_4a : str
    chromosome_4b : str
    location : str
        A string describing where this stock is located.
    created_date : date
        The date at which this stock was created.
    source : str
        The source of this stock. This might be either the token of the person
        that created this stock or Bloomington, VDRC, Kyoto, etc.
    documentation : str
        If this stock was created in our lab this should be a short
        description about how this stock was created or, even better,
        the pages in the lab book describing the generation of this stock.
    reference : str
        The DOI of a publication in which this stock was used.
    discarded_date : date
        The date at which this stock was discarded or died out.

    """

    id: Mapped[int] = mapped_column(
        ForeignKey("base_entity.id"), primary_key=True, info={"importable": False}
    )
    short_genotype: Mapped[str] = mapped_column(
        String(2048), nullable=True, info={"importable": True}
    )
    chromosome_xa: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_xb: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_y: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_2a: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_2b: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_3a: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_3b: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_4a: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_4b: Mapped[str] = mapped_column(
        String(2048), nullable=False, default="+", info={"importable": True}
    )
    location: Mapped[str] = mapped_column(String(64), nullable=True, info={"importable": True})
    created_date: Mapped[date] = mapped_column(CustomDate, nullable=True, info={"importable": True})
    source: Mapped[str] = mapped_column(String(512), nullable=True, info={"importable": True})
    documentation: Mapped[str] = mapped_column(
        String(2048), nullable=True, info={"importable": True}
    )
    reference: Mapped[str] = mapped_column(String(512), nullable=True, info={"importable": True})
    discarded_date: Mapped[date] = mapped_column(
        CustomDate, nullable=True, info={"importable": True}
    )

    # One-to-many relationships.
    modifications: Mapped[list["Modification"]] = relationship(
        backref="fly_stock", order_by="Modification.date.desc()", lazy=True
    )

    # Proper setup for joined table inheritance.
    __mapper_args__ = {"polymorphic_identity": "fly_stock"}

    @classmethod
    def _filters(cls, **fields) -> list:
        filters = []

        match fields.pop("discarded", "all"):
            case "discarded":
                filters.append(cls.discarded_date.isnot(None))
            case "recent":
                filters.append(cls.discarded_date.is_(None))
            case _:
                pass

        return super()._filters(**fields) + filters



import pytest
from flask import url_for
from flask_login import current_user, login_user
from sqlalchemy import func, select

from labbase2 import models
from labbase2.database import db


def test_login_get(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.get(url)

    assert response.status_code == 200
    assert b"Sign in" in response.data


def test_login_with_wrong_user(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "wrong_email@test.de", "password": "admin", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Invalid email address or username!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_login_with_wrong_pw(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "test@test.de", "password": "wrong_pw", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Wrong password!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_login_inactive_user(app, client):
    with app.app_context(), client:
        user = db.session.get(models.User, 1)
        user.is_active = False

        db.session.commit()

        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "test@test.de", "password": "admin", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Your account is inactive!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_login_with_correct_data(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "test@test.de", "password": "admin", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Successfully logged in!" in response.data
        assert not current_user.is_anonymous
        assert current_user.is_authenticated
        assert current_user.username == "Max Mustermann"


def test_logout_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        url = url_for("auth.logout")
        user = db.session.get(models.User, 1)
        login_user(user)
        response = client.get(url, follow_redirects=True)

        assert b"Successfully logged out!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_register_existing_email(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        login_user(admin)

        url = url_for("auth.register")
        response = client.post(
            url,
            data={
                "first_name": "Maja",
                "last_name": "Musterfrau",
                "email": "test@test.de",
                "password": "This_isAPassword123",
                "password2": "This_isAPassword123",
                "submit": True,
            },
            follow_redirects=True,
        )

        user_count = db.session.scalar(select(func.count()).select_from(models.User))

        assert b"Email address already exists!" in response.data
        assert user_count == 1


def test_register_new_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        login_user(admin)

        url = url_for("auth.register")
        response = client.post(
            url,
            data={
                "first_name": "Maja",
                "last_name": "Musterfrau",
                "email": "test2@test.de",
                "password": "This_isAPassword123",
                "password2": "This_isAPassword123",
                "submit": True,
            },
            follow_redirects=True,
        )

        user_count = db.session.scalar(select(func.count()).select_from(models.User))
        new_user = db.session.get(models.User, 2)
        user_group = db.session.get(models.Group, "user")

        assert b"Successfully registered new user!" in response.data
        assert user_count == 2
        assert user_group in new_user.groups


def test_get_edit_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.edit_user")
        response = client.get(url)

        assert response.status_code == 200
        assert b"Edit Profile" in response.data


def test_edit_user_with_wrong_pw(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.edit_user")
        response = client.post(
            url,
            data={
                "first_name": "Maximilian",
                "last_name": "Mustermann",
                "email": "test@test.de",
                "timezone": "UTC",
                "password": "wrong_password",
            },
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Password is incorrect!" in response.data
        assert user.first_name == "Max"


def test_edit_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.edit_user")
        response = client.post(
            url,
            data={
                "first_name": "Maximilian",
                "last_name": "Mustermann",
                "email": "test@test.de",
                "timezone": "UTC",
                "password": "admin",
            },
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Successfully edited user profile!" in response.data
        assert user.first_name == "Maximilian"


def test_remove_only_admin(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.change_admin_status", id_=1)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed admin setting for" in response.data
        assert b"No user with admin rights! Reverting previous change!" in response.data
        assert user.is_admin


def test_make_user_admin(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        group = db.session.get(models.Group, "admin")
        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        login_user(admin)
        url = url_for("auth.change_admin_status", id_=2)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed admin setting for" in response.data
        assert b"No user with admin rights! Reverting previous change!" not in response.data
        assert group in admin.groups


def test_remove_admin_from_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        group = db.session.get(models.Group, "admin")
        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.groups.append(group)
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        login_user(admin)
        url = url_for("auth.change_admin_status", id_=2)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed admin setting for" in response.data
        assert b"No user with admin rights! Reverting previous change!" not in response.data
        assert group not in user.groups


# TODO Test auth.change_password
# TODO Test auth.change_permissions
# TODO Test auth.change_active_status


def test_inactivate_only_admin(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        login_user(admin)

        assert admin.is_active

        url = url_for("auth.change_active_status", id_=1)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed active setting for" in response.data
        assert b"No active user! Reverting previous change!" in response.data
        assert admin.is_active


def test_inactivate_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        login_user(admin)

        assert user.is_active

        url = url_for("auth.change_active_status", id_=2)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed active setting for" in response.data
        assert not user.is_active


# TODO Test auth.create_password_reset


def test_users_get(app, client):
    with app.app_context(), app.test_request_context(), client:
        url = url_for("auth.users")

        admin = db.session.get(models.User, 1)
        login_user(admin)

        response = client.get(url)

        assert response.status_code == 200
        assert b"Mustermann" in response.data
        assert b"Musterfrau" not in response.data

        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        response = client.get(url)

        assert response.status_code == 200
        assert b"Mustermann" in response.data
        assert b"Musterfrau" in response.data


def test_add_edit_group(app, client):
    with app.app_context(), app.test_request_context(), client:
        url = url_for("auth.add_group")

        admin = db.session.get(models.User, 1)
        login_user(admin)

        # Check site is accessible.
        response = client.get(url)
        assert response.status_code == 200
        assert b"Add Group</h1>" in response.data

        # Check group can be added.
        response = client.post(
            url,
            data={"name": "students", "description": "Fake group.", "p_edit_groups": True},
            follow_redirects=True,
        )
        group = db.session.get(models.Group, "students")
        permission = db.session.get(models.Permission, "edit-groups")
        assert response.status_code == 200
        assert b"Successfully added group" in response.data
        assert group is not None
        assert group.description == "Fake group."
        assert permission in group.permissions

        # Check site is accessible.
        url = url_for("auth.edit_group", id_="students")
        response = client.get(url)
        assert response.status_code == 200
        assert b"Edit Group" in response.data

        response = client.post(
            url,
            data={"description": "No fake group."},
            follow_redirects=True,
        )
        assert response.status_code == 200
        assert group.description == "No fake group."
        assert permission not in group.permissions


def test_delete_admin_user_group(app):
    with app.app_context():
        admin_group = db.session.get(models.Group, "admin")
        user_group = db.session.get(models.Group, "user")

        db.session.delete(admin_group)

        with pytest.raises(ValueError, match="Groups 'admin' and 'user' cannot be deleted!"):
            db.session.commit()

        db.session.rollback()
        db.session.delete(user_group)

        with pytest.raises(ValueError, match="Groups 'admin' and 'user' cannot be deleted!"):
            db.session.commit()

        db.session.rollback()
        groups = db.session.scalars(select(models.Group.name)).all()

        assert "admin" in groups
        assert "user" in groups


def test_remover_user_membership(app, client):
    with app.app_context():
        user = db.session.get(models.User, 1)
        user_group = db.session.get(models.Group, "user")
        admin_group = db.session.get(models.Group, "admin")

        with pytest.raises(ValueError, match="Group 'user' cannot be removed from a user!"):
            user.groups.remove(user_group)
            db.session.commit()

        db.session.rollback()

        with pytest.raises(ValueError, match="Group 'user' cannot be removed from a user!"):
            user.groups = [admin_group]
            db.session.commit()


def test_change_groups(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        new_group = models.Group(name="students")
        db.session.add(new_group)
        db.session.commit()

        url = url_for("auth.change_groups", id_=1)

        response = client.get(url)
        assert response.status_code == 200
        assert b"Edit Groups for user" in response.data

        response = client.post(url, data={"g_students": True}, follow_redirects=True)
        assert response.status_code == 200
        assert new_group in user.groups



import secrets
from datetime import datetime, timedelta
from typing import Optional
from zoneinfo import ZoneInfo

from email_validator import validate_email
from flask import Blueprint, Response
from flask import current_app as app
from flask import flash, redirect, render_template, request, url_for
from flask_login import current_user, login_required, login_user, logout_user
from sqlalchemy import func, select
from sqlalchemy.exc import DataError

from labbase2.database import db
from labbase2.models import BaseFile, Group, Permission, ResetPassword, User
from labbase2.utils.permission_required import permission_required
from labbase2.views.files.routes import upload_file

from .forms.edit import EditUserForm
from .forms.edit_permissions import EditPermissions
from .forms.groups import AddGroupForm, ChangeGroupsForm, EditGroupForm
from .forms.login import LoginForm
from .forms.password import ChangePassword
from .forms.register import RegisterForm

__all__ = ["bp", "login"]


# The blueprint to register all coming blueprints with.
bp = Blueprint("auth", __name__, url_prefix="/auth", template_folder="templates")


# @bp.route("/", methods=["GET"])
# @login_required
# def index():
#     return render_template("user/index.html", user=current_user)


@bp.route("/login", methods=["GET", "POST"])
def login() -> str | Response:
    # Current user is already authenticated.
    if current_user.is_authenticated:
        flash("You are already logged in!", "warning")
        return redirect(url_for("base.index"))

    form = LoginForm()

    # POST fork with correct credentials.
    if form.validate_on_submit():
        user = db.session.scalar(select(User).where(User.email == form.email.data))

        if not user:
            app.logger.warning("Login attempt with invalid credentials")
            flash("Invalid email address or username!", "danger")
        elif not user.is_active:
            app.logger.info("Deactivate user tried to log in: %s", user.username)
            flash("Your account is inactive! Get in touch with an admin!", "warning")
        elif not user.verify_password(form.password.data):
            app.logger.warning("Login attempt with invalid password: %s", user.username)
            flash("Wrong password!", "danger")
        else:
            login_user(user, remember=form.remember_me.data)
            app.logger.info("Logged in: %s", user.username)
            flash("Successfully logged in!", "success")

            last_login = current_user.timestamp_last_login
            if last_login is None:
                flash("This is your first log in!", "info")
            else:
                tz = current_user.timezone
                formatted = last_login.astimezone(ZoneInfo(tz)).strftime("%B %d, %Y %I:%M %p")
                flash(f"Last login was on {formatted}!", "info")

            user.timestamp_last_login = func.now()  # pylint: disable=not-callable
            try:
                db.session.commit()
            except Exception as error:
                db.error("Unable to save login time due to unknown database error: %s", error)

            next_page = request.args.get("next")

            if not next_page:  # or url_parse(next_page).netloc:
                next_page = url_for("base.index")

            return redirect(next_page)

    # GET fork or POST with wrong credentials.
    return render_template("auth/login.html", title="Login", login_form=form)


@bp.route("/logout", methods=["GET"])
@login_required
def logout() -> str | Response:
    app.logger.info("Logged out.")
    flash("Successfully logged out!", "success")

    logout_user()

    return redirect(url_for(".login"))


@bp.route("/register", methods=["GET", "POST"])
@login_required
@permission_required("register-user")
def register():
    form = RegisterForm()

    # POST fork of view.
    if form.validate_on_submit():
        email = validate_email(form.email.data).normalized

        user_count_stm = (
            select(func.count())  # pylint: disable=not-callable
            .select_from(User)
            .where(User.email == email)
        )
        if db.session.scalar(user_count_stm) > 0:
            app.logger.warning("Tried to register new user with existing email address: %s", email)
            flash("Email address already exists!", "danger")
        else:
            user = User()
            form.populate_obj(user)
            user.email = email
            user.set_password(form.password.data)

            try:
                db.session.add(user)
                db.session.commit()
            except DataError as error:
                app.logger.warning("Could not write new user to database: %s", error)
                flash(str(error), "danger")
                db.session.rollback()
            else:
                app.logger.info("Registered new user: %s", user.username)
                flash("Successfully registered new user!", "success")

    elif request.method.lower() == "post":
        app.logger.warning("Could not register new user due to invalid input: %s", form.data)
        for field, errors in form.errors.items():
            for error in errors:
                flash(f"{field}: {error}", "danger")

    return render_template("auth/register.html", title="Register", form=form)


@bp.route("/edit", methods=["GET", "POST"])
@login_required
def edit_user():
    if request.method == "GET":
        form = EditUserForm(None, obj=current_user)
        return render_template("auth/edit-user.html", title="Edit profile", form=form)

    form = EditUserForm()

    if not form.validate():
        app.logger.warning("Could not edit user due to invalid input.")
        for error in form.errors:
            flash(str(error), "danger")
        return render_template("auth/edit-user.html", title="Edit profile", form=form)

    if not current_user.verify_password(form.password.data):
        app.logger.warning("Entered incorrect password.")
        flash("Password is incorrect!", "danger")
    else:
        form.populate_obj(current_user)

        if form.file.data:
            file = upload_file(form, BaseFile)
            current_user.picture = file
            file.resize(512)

        try:
            db.session.commit()
        except Exception as error:
            app.logger.warning("Could not update user: %s", error)
            flash(str(error), "error")
            db.session.rollback()
        else:
            app.logger.info("Updated user profile.")
            flash("Successfully edited user profile!", "success")

    return render_template("auth/edit-user.html", title="Edit profile", form=form)


@bp.route("/change-password", methods=["GET", "POST"], defaults={"key": None})
@bp.route("/change-password/<string:key>", methods=["GET", "POST"])
@login_required
def change_password(key: Optional[str] = None):
    if key is not None:
        reset = db.session.get(ResetPassword, key)
        if not reset:
            flash("Invalid token!", "danger")
            return redirect(url_for("auth.change_password"))
        if datetime.now() > reset.timeout:
            flash("Token to reset password has expired!")
            db.session.delete(reset)
            return redirect(url_for("auth.change_password"))
        user = reset.user
        logout_user()
        login_user(user)
    else:
        reset = None
        user = current_user

    form = ChangePassword()

    if request.method == "GET":
        return render_template("auth/change-password.html", form=form)

    if not form.validate():
        flash(f"{form.errors}", "danger")

    if not reset and not current_user.verify_password(form.old_password.data):
        flash("Old password incorrect!", "danger")
        render_template("auth/change-password.html", form=form)

    user.set_password(form.new_password.data)
    user.resets.clear()

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        flash(str(error), "danger")
    else:
        flash("Your password has been updated!", "success")

    return render_template("auth/change-password.html", form=form)


@bp.route("/change-permissions", methods=["GET"], defaults={"id_": None})
@bp.route("/change-permissions/<int:id_>", methods=["GET", "POST"])
@login_required
@permission_required("change-group")
def change_permissions(id_: int = None):
    users_stmt = select(User).where(User.is_active).order_by(User.last_name, User.first_name)

    if id_ is None:
        user = current_user
    else:
        user = db.session.get(User, id_)

    form = EditPermissions(**user.form_permissions)

    if form.validate_on_submit():
        user.permissions.clear()
        for field in form:
            if not field.type == "BooleanField":
                continue
            if not field.data:
                continue

            permission = db.session.get(Permission, field.name.replace("_", " ").capitalize())
            user.permissions.append(permission)

        db.session.commit()
        app.logger.info("Updated permissions for user: %s", user.username)
        flash(f"Successfully updated permissions for {user.username}!", "success")

    return render_template(
        "auth/edit-permissions.html",
        form=form,
        selected_user=user,
        users=db.session.scalars(users_stmt).all(),
    )


@bp.route("/change-admin-status/<int:id_>", methods=["GET"])
@login_required
@permission_required("change-group")
def change_admin_status(id_: int):
    user = db.session.get(User, id_)
    admin_group = db.session.get(Group, "admin")

    if user is None:
        flash(f"No user with ID {id_})!", "danger")
        return redirect(url_for(".users"))

    if admin_group in user.groups:
        user.groups.remove(admin_group)
    else:
        user.groups.append(admin_group)

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        flash(str(error), "danger")
    else:
        flash(f"Successfully changed admin setting for {user.username}!", "success")

    user_count_stm = (
        select(func.count())  # pylint: disable=not-callable
        .select_from(User)
        .join(User.groups)
        .where(Group.name == "admin")
    )
    if db.session.scalar(user_count_stm) == 0:
        flash("No user with admin rights! Reverting previous change!", "warning")
        user.groups.append(admin_group)
        db.session.commit()

    return redirect(url_for(".users"))


@bp.route("/change-active-status/<int:id_>", methods=["GET"])
@login_required
@permission_required("inactivate-user")
def change_active_status(id_: int):
    user = db.session.get(User, id_)

    if user is None:
        flash(f"No user with ID {id_})!", "danger")
    else:
        user.is_active = not user.is_active

        try:
            db.session.commit()
        except Exception as error:
            db.session.rollback()
            flash(str(error), "danger")
        else:
            flash(f"Successfully changed active setting for {user.username}!", "success")

    user_count_stm = (
        select(func.count()).select_from(User).where(User.is_active)  # pylint: disable=not-callable
    )
    if db.session.scalar(user_count_stm) == 0:
        flash("No active user! Reverting previous change!", "warning")
        user.is_active = True
        db.session.commit()

    if request.referrer is None:
        return redirect(url_for(".users"))

    return redirect(request.referrer)


@bp.route("/reset-password/<int:id_>", methods=["GET"])
@login_required
@permission_required("reset-password")
def create_password_reset(id_: int):
    user = db.session.get(User, id_)

    if user is None:
        flash(f"No user with ID {id_})!", "danger")
        return redirect(request.referrer)

    user.resets.clear()

    db.session.commit()

    expires_after = app.config.get("RESET_EXPIRES", 10)
    reset = ResetPassword(
        token=secrets.token_urlsafe(32),
        user_id=id_,
        timeout=datetime.now() + timedelta(minutes=expires_after),
    )

    db.session.add(reset)
    db.session.commit()

    flash(f"Generated a password reset . Expires: {reset.timeout.isoformat()}!", "success")
    flash(f"Visit: {url_for('auth.change_password', key=reset.token)}")

    if request.referrer is None:
        return redirect(url_for(".users"))

    return redirect(request.referrer)


@bp.route("/users", methods=["GET"])
@login_required
@permission_required("view-user")
def users():
    entities = db.session.scalars(
        select(User).order_by(User.is_active.desc(), User.last_name, User.first_name)
    )
    return render_template("auth/users.html", users=entities, title="Users")


@bp.route("/groups", methods=["GET"])
@login_required
@permission_required("edit-groups")
def index_groups():
    entities = db.session.scalars(
        select(Group).where(Group.name.notin_(["admin", "user"])).order_by(Group.name.asc())
    )

    return render_template("auth/groups.html", groups=entities, title="Groups")


@bp.route("/groups/add", methods=["GET", "POST"])
@login_required
@permission_required("edit-groups")
def add_group():
    form = AddGroupForm()

    if request.method == "GET":
        return render_template("auth/add-group.html", form=form)

    # This is required as fields for permissions are dynamically added.
    form.process(formdata=request.form)

    if not form.validate():
        app.logger.info("Couldn't add group due to invalid input.")
        flash(f"{form.errors}", "warning")

    group = Group()
    group.name = form.name.data
    group.description = form.description.data
    group.permissions = form.selected_permissions

    try:
        db.session.add(group)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning("Couldn't add group due to unknown database error: %s", error)
        flash(str(error), "warning")
    else:
        flash(f"Successfully added group '{group.name}' to database.", "success")

    return render_template("auth/add-group.html", form=form)


@bp.route("/groups/<string:id_>", methods=["GET", "POST"])
@login_required
@permission_required("edit-groups")
def edit_group(id_: str):
    if not (group := db.session.get(Group, id_)):
        flash(f"No group with name {id_}!", "warning")
        return redirect(request.referrer)

    if group.name in ["user", "admin"]:
        flash("'user' and 'admin' groups cannot be edited!", "warning")
        return redirect(request.referrer)

    form = EditGroupForm(obj=group, group_name=id_)

    if request.method == "GET":
        return render_template("auth/edit-group.html", form=form, group=group)

    # This is required as fields for permissions are dynamically added.
    form.process(formdata=request.form)

    if not form.validate():
        flash(f"{form.errors}", "warning")
        return render_template("auth/edit-group.html", form=form, group=group)

    group.description = form.description.data
    group.permissions = form.selected_permissions

    try:
        db.session.commit()
    except Exception as error:
        flash(f"Could not edit group: {error}", "warning")
        db.session.rollback()
    else:
        flash(f"Successfully edited group {group.name}!", "success")

    return render_template("auth/edit-group.html", form=form, group=group)


@bp.route("users/groups/<int:id_>", methods=["GET", "POST"])
@login_required
@permission_required("change-group")
def change_groups(id_: int):
    if not (user := db.session.get(User, id_)):
        flash(f"No user with ID {id_}!", "warning")
        return redirect(request.referrer)

    form = ChangeGroupsForm(user_id=id_)

    if request.method == "GET":
        return render_template("auth/change-groups.html", form=form, user=user)

    # Required because fields are dynamically added to the form.
    form.process(formdata=request.form)

    if not form.validate():
        flash(f"{form.errors}", "warning")
        return render_template("auth/change-groups.html", form=form, user=user)

    protected_groups = db.session.scalars(
        select(Group).where(Group.name.in_(["admin", "user"]), Group.users.any(User.id == id_))
    ).all()

    user.groups = form.selected_groups + protected_groups

    try:
        db.session.commit()
    except Exception as error:
        flash(f"Could not change groups! {error}", "warning")
        db.session.rollback()
    else:
        flash("Successfully changed groups!", "success")

    return render_template("auth/change-groups.html", form=form, user=user)



import io
from datetime import date

from Bio import SeqIO
from Bio.Restriction import CommOnly
from Bio.SeqRecord import SeqRecord
from flask import Response, send_file

__all__ = ["Sequence"]


class Sequence:
    """A mixin to export sequence entities (primers/plasmids) to FASTA files"""

    def __len__(self):
        return len(self.seqrecord.seq)

    @property
    def seqrecord(self) -> SeqRecord:
        """Generate a Biopython SeqRecord object

        Returns
        -------
        SeqRecord
            The instance of this class represented as a Biopython SeqRecord object.
        """

        raise NotImplementedError

    def restriction_sites(self, sites: int = 1) -> dict["RestrictionType", int]:
        """Return a list of cutting restriction enzymes

        Parameters
        ----------
        sites: int, optional
            The maximum number of sites a given restriction enzymes cuts. Set to 0 to
            get all cutting enzymes.

        Returns
        -------
        dict[RestrictionType, int]
            A list of all restriction enzymes that cut the sequence at most the
            specified times but at least once.
        """

        if record := self.seqrecord:
            enzymes = CommOnly.search(record.seq, linear=False)
            enzymes = {k: len(v) for k, v in enzymes.items() if len(v) <= sites or sites == 0}

            return enzymes

        return {}

    def formatted_restriction_sites(self) -> str:
        """Return restriction sites with HTML markup.

        Returns
        -------
        str
            A single string that can be inserted into a website as it is. It
            highlights the number of cutting sites.
        """

        sites = list(self.restriction_sites().items())
        sites.sort(key=lambda x: (x[1], x[0]))

        return ", ".join([f"<b>{e}</b> ({n})" for e, n in sites])

    @classmethod
    def to_fasta(cls, instances: list) -> Response:
        """Export instances to a file with multiple FASTA entries

        Parameters
        ----------
        instances: list
            A list of instances to export to a single FASTA file.

        Returns
        -------
        Response
            A file for download.
        """

        with io.StringIO() as proxy:
            SeqIO.write([i.seqrecord for i in instances], proxy, "fasta")
            mem = io.BytesIO(proxy.getvalue().encode("utf-8"))

        filename = cls.__name__ + "_" + date.today().isoformat() + ".fasta"

        return send_file(mem, as_attachment=True, download_name=filename, mimetype="text/fasta")



from datetime import datetime

from sqlalchemy import DateTime, ForeignKey, String, func
from sqlalchemy.orm import Mapped, mapped_column

from labbase2.database import db
from labbase2.models.mixins.export import Export
from labbase2.models.mixins.importer import Importer

__all__ = ["Comment"]


class Comment(db.Model, Importer, Export):
    """A comment to an entity.

    Attributes
    ----------
    id : int
        The identifier of this comment.
    entity_id : int
        The identifier of the entity about which this comment is.
    user_id : str
        The id of the person that has written the comment.
    timestamp_created : DateTime
        The time of the comment. This is automatically set by the database.
    timestamp_edited : DateTime
        The time this comment was last edited. Might be `None` if the comment was
        never modified.
    subject : str
        A short string describing the subject of the comment.
    text : str
        The message of the comment.
    """

    __tablename__: str = "comment"

    id: Mapped[int] = mapped_column(primary_key=True, info={"importable": False})
    entity_id: Mapped[int] = mapped_column(
        ForeignKey("base_entity.id"), nullable=False, info={"importable": True}
    )
    user_id: Mapped[int] = mapped_column(
        ForeignKey("user.id"), nullable=False, info={"importable": True}
    )
    timestamp_created: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),  # pylint: disable=not-callable
        info={"importable": True},
    )
    timestamp_edited: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=True,
        onupdate=func.now(),  # pylint: disable=not-callable
        info={"importable": True},
    )
    subject: Mapped[str] = mapped_column(String(128), nullable=False, info={"importable": True})
    text: Mapped[str] = mapped_column(String(2048), nullable=False, info={"importable": True})

    __table_args__ = {"extend_existing": True}



from functools import wraps
from typing import Callable

from flask import current_app as app
from flask import flash, redirect, url_for
from flask_login import current_user

from labbase2.database import db
from labbase2.models import Permission

__all__ = ["permission_required"]


def permission_required(*allowed) -> Callable:
    """Check whether the current user has sufficient role to access a resource.

    This is a very simple decorator op to add user role system to the
    website.

    Parameters
    ----------
    *allowed : *str
        A list of permissions. If the user has any of these permissions, he will be
        granted access to the view.

    Returns
    -------
    function
        The decorate route function.
    """

    def decorator(func: Callable):

        @wraps(func)
        def decorated_view(*args, **kwargs):

            if current_user.is_admin:
                return func(*args, **kwargs)

            for name in allowed:
                permission = db.session.get(Permission, name)
                if not permission:
                    app.logger.debug("No such permission: %s", name)
                elif current_user.has_permission(permission):
                    return func(*args, **kwargs)

            flash("No permission to enter this site!", "warning")
            return redirect(url_for("base.index"))

        return decorated_view

    return decorator



from flask_wtf import FlaskForm
from wtforms.fields import FileField, StringField, SubmitField, TextAreaField
from wtforms.validators import DataRequired, Length, Optional

from labbase2.forms import rendering
from labbase2.forms.filters import strip_input

__all__ = ["UploadFile"]


class UploadFile(FlaskForm):
    """Form to upload files.

    Attributes
    ----------

    """

    file = FileField("Select File", validators=[DataRequired()], render_kw=rendering.file_field)
    filename = StringField(
        "Filename",
        validators=[Optional(), Length(max=64)],
        render_kw=rendering.custom_field | {"placeholder": "(Optional)"},
        description="Choose an optional filename.",
    )
    note = TextAreaField(
        "Note",
        validators=[Optional(), Length(max=2048)],
        filters=[strip_input],
        render_kw=rendering.custom_field | {"placeholder": "Note", "rows": 8},
    )
    submit = SubmitField(label="Submit", render_kw=rendering.submit_field)



from flask import Blueprint
from flask_login import current_user, login_required

from labbase2.database import db
from labbase2.forms.utils import errors2messages
from labbase2.models import Modification

from .forms import EditModification

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint("modifications", __name__, url_prefix="/modifications", template_folder="templates")


@bp.route("/<int:flystock_id>", methods=["POST"])
@login_required
def add(flystock_id: int):
    if (form := EditModification()).validate():
        modification = Modification(fly_id=flystock_id, user_id=current_user.id)
        form.populate_obj(modification)

        try:
            db.session.add(modification)
            db.session.commit()
        except Exception as error:
            return str(error), 400

        return "Successfully added modification!", 201

    return errors2messages(form.errors), 400


@bp.route("/<int:id>", methods=["PUT"])
@login_required
def edit(id_: int):
    if (form := EditModification()).validate():
        if (modification := db.session.get(Modification, id_)) is None:
            return f"No preparation with ID {id_}!", 404
        if modification.user_id != current_user.id:
            return "Modification can only be edited by source user!", 400

        form.populate_obj(modification)

        try:
            db.session.commit()
        except Exception as error:
            return str(error), 400

        return "Successfully edited modification!", 200

    return errors2messages(form.errors), 400


@bp.route("/<int:flystock_id>/<int:id_>", methods=["DELETE"])
@login_required
def delete(flystock_id: int, id_: int):
    if (modification := db.session.get(Modification, id_)) is None:
        return f"No modification with ID {id}!", 404
    if modification.user_id != current_user.id:
        return "Modification can only be deleted by source user!", 400

    try:
        db.session.delete(modification)
        db.session.commit()
    except Exception as error:
        return str(error), 400

    return f"Successfully deleted modification {id}!", 200
